package swarmnet;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class Network {

	private List<Node> nodes;
	private Random random;
	private Node sink;

	public Network(int width, int height, int numNodes, double neighbourDistance,
			Random r) {

		random = r;
		initNodes(width, height, numNodes, neighbourDistance);
	}

	public void initNodes(int width, int height, int numNodes, double neighbourDistance) {

		nodes = new ArrayList<Node>();

		// Create the sink as first element of nodes list
		sink = new Node(3 * width / 4, 3 * height / 4);
		nodes.add(sink);

		// Create nodes at random
		for (int i = 1; i < numNodes; ++i) {

			Node n = makeRandomNode(width, height);
			nodes.add(n);
		}

		// Each node knows its neighbours i.e. those within hopLength
		for (Node n : nodes) {
			n.setNeighbours(nodes, neighbourDistance);
		}

		// Set the hop distance of each node from the sink
		setHops(neighbourDistance);

		// Some nodes may be unreachable from the sink
		int numRemoved = 0;
		for (int i = 0; i < nodes.size(); ++i) {
			Node n = nodes.get(i);
			if (n.getHopsToSink() == -1) {
				nodes.remove(i);
				++numRemoved;
				--i;
			}
		}

		// Remove unreachable nodes and generate new nodes within one hop of a connected node
		for (int i = 0; i < numRemoved; ++i) {

			Node randNode = nodes.get(random.nextInt(nodes.size()));
			double r = random.nextDouble() * neighbourDistance;
			double theta = random.nextDouble() * 2 * Math.PI;
			int x = randNode.getX() + (int) (r * Math.cos(theta));
			int y = randNode.getY() + (int) (r * Math.sin(theta));

			Node n = new Node(x, y);
			nodes.add(n);
			n.setNeighbours(nodes, neighbourDistance);
			List<Node> neighbours = n.getNeighbours();
			int hop = Integer.MAX_VALUE;
			for (Node neighbour : neighbours)
				hop = Math.min(hop, neighbour.getHopsToSink() + 1);
			n.setHopsToSink(hop);
		}
	}

	public Node makeRandomNode(int w, int h) {

		boolean tryAgain = true;
		int x = -1, y = -1;
		while (tryAgain) {

			x = random.nextInt(w);
			y = random.nextInt(h);

			for (Node n : nodes)
				if (n.getX() != x && n.getY() != y)
					tryAgain = false;
		}

		return new Node(x, y);
	}

	
	private void setHops(double neighbourDistance) {

		sink.setHopsToSink(0);
		boolean finish = false;
		int h = 1;
		while (!finish) {
			int numProcessed = 0;
			for (Node nodeA : nodes) {
				if (nodeA.getHopsToSink() != -1)
					continue;

				for (Node nodeB : nodes) {

					if (nodeB.equals(nodeA))
						continue;

					if (nodeB.getHopsToSink() == h - 1
							&& Node.dist(nodeA, nodeB) < neighbourDistance) {
						nodeA.setHopsToSink(h);
						++numProcessed;
					}
				}

			}
			if (numProcessed == 0)
				finish = true;
			h++;
		} // end while
	}

	public List<Node> getNodes() {
		return nodes;
	}

	public Node getSink() {
		return sink;
	}
}
