package swarmnet;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class SwarmNet {

    private Network network;
    private List<Node> nodes;
    private List<Node> nodesWithPackets;
    private List<Node> closerNodes, notCloserNodes, attrNodes;
    private Random random;
    private Node sink;
    private int numPackets;
    private int maxQueue;
    private int numNodes;

    private int numMoves;
    private int numAttempts;

    private double neighbourDistance;

    private int algo;
    private double hopWeight, mostAttrQ;

    private int width, height;

    public SwarmNet(int w, int h, long configSeed, long randomSeed, int a) {

        width = w;
        height = h;

        Random configRandom = new Random(configSeed);
        random = new Random(randomSeed);
        initNetwork(configRandom);
        initPackets();

        algo = a;
        hopWeight = 2.0;
        mostAttrQ = maxQueue / 4.0;
    }

    void initNetwork(Random configRandom) {

        maxQueue = 10;
        numNodes = 500;
        neighbourDistance = 50;

        numMoves = numAttempts = 0;

        network = new Network(width, height, numNodes, neighbourDistance, configRandom);
        nodes = network.getNodes();
        sink = network.getSink();

        closerNodes = new ArrayList<Node>();
        notCloserNodes = new ArrayList<Node>();
        attrNodes = new ArrayList<Node>();
    }

    public void initPackets() {

        numPackets = 0;
        for (int i = 0; i < numNodes; ++i) {

            Node node = nodes.get(i);
            if ((node.getX() < width / 4 && node.getY() < width / 4)) {

                
                for (int j = 0; j < maxQueue; ++j) {

                    Packet p = new Packet(node);
                    node.addPacket(p);
                    ++numPackets;
                }
            }
        }
        if (numPackets == 0) {
            exit("Could not distribute packets...exiting...");
        }
        nodesWithPackets = new ArrayList<Node>();
        for (Node n : nodes) {
            if (n.numPackets() > 0)
                nodesWithPackets.add(n);
        }
    }
    
public void movePacket(Node fromNode, Node toNode, Packet p) {

        ++numMoves;

        p.setPreviousNode(fromNode);
        fromNode.removePacket(p);

        if (fromNode.numPackets() == 0)//&& nodesWithPackets.contains(fromNode))
            nodesWithPackets.remove(fromNode);

        p.setNode(toNode);
        toNode.addPacket(p);

        if (!nodesWithPackets.contains(toNode))
            nodesWithPackets.add(toNode);
    }

    private Node randomNodeInList(List<Node> l) {
        return l.get(random.nextInt(l.size()));
    }

    // Algorithm A
    // 1. If any neighbour is sink, definitely send packet there.
    // 2. Choose at random amongst any closer nodes that have spare capacity. (This may be the packet's previously visited node.)
    // 3. If there are no closer available nodes, find the most attractive of available nodes at h = 0 and h = -1, 
    //    omitting the node previously visited by this packet.
    // 4. If step 3 doesn't find a node, then send to previously visited node, if this is available.
    public Node algorithmA(Node node, Packet packet) {

        Node prefNode = node;

        closerNodes.clear();
        notCloserNodes.clear();

        for (Node n : node.getNeighbours()) {

            if (n.equals(sink)) {
                return n;
            }
            if (n.numPackets() < maxQueue) {

                int hops = node.getHopsToSink() - n.getHopsToSink();
                if (hops > 0) {
                    closerNodes.add(n);
                } else {
                    notCloserNodes.add(n);
                }
            }
        }

        if (closerNodes.size() > 0) {
            prefNode = randomNodeInList(closerNodes);
        } else if (notCloserNodes.size() > 0) {

            attrNodes.clear();

            long currAttr = Long.MIN_VALUE;
            for (int i = 0; i < notCloserNodes.size(); ++i) {

                Node n = notCloserNodes.get(i);

                if (packet.getPreviousNode().equals(n))
                    continue;

                long attr = Math.round(mostAttrQ
                        - Math.abs(n.numPackets() - mostAttrQ));

                if (attr == currAttr)
                    attrNodes.add(n);

                else if (attr > currAttr) {
                    attrNodes.clear();
                    attrNodes.add(n);
                    currAttr = attr;
                }

            }
            if (!attrNodes.isEmpty())
                prefNode = randomNodeInList(attrNodes);
            else
                prefNode = packet.getPreviousNode();
        } else {
            // All neighbouring nodes are full.
            // Return input node. 
        }

        return prefNode;
    }

    // Algorithm B
    // 1. If neighbour is sink, go there.
    // 2. Find most attractive available neighbour according to 
    // attraction = hopWeight * hops + mostAttractiveQueueLength - |deltaPackets|
    // where deltaPackets is difference between packets on neighbour and most attractive queue length
    // and hops is hop difference between this node and its neighbour i.e. hops = 1, 0 or -1.
    // If there is a tie, then the most attractive is chosen at random from the winning neighbours.
    // 3. If no neighbour is attractive, return to previous node, if this is available.
    public Node algorithmB(Node node, Packet packet) {

        attrNodes.clear();
        int currAttr = Integer.MIN_VALUE;
        List<Node> neighbours = node.getNeighbours();
        for (Node n : neighbours) {

            if (packet.getPreviousNode().equals(n))
                continue;

            if (n.equals(sink))
                return n;

            if (n.numPackets() == maxQueue)
                continue;

            int hops = node.getHopsToSink() - n.getHopsToSink();

            int attr = (int)Math.round(hopWeight * hops + mostAttrQ
                    - Math.abs(n.numPackets() - mostAttrQ));

           
//           int attr = (int)Math.round(2 * hops + maxQueue / 4.0 - n.numPackets());

            if (attr == currAttr)
                attrNodes.add(n);

            else if (attr > currAttr) {
                attrNodes.clear();
                attrNodes.add(n);
                currAttr = attr;
            }

        }

        if (!attrNodes.isEmpty())
            return randomNodeInList(attrNodes);
        else if (packet.getPreviousNode().numPackets() < maxQueue)
            return packet.getPreviousNode();

        return node;
    }

    // From empirical tests using the ns-2 network simulator:
    // max Q length = 50
    // pks generated at 25, 35 or 45 pks/sec from a few active nodes in close proximity.
    // Measure packet delivery rate (PDR) and end to end delay (EED).
    public void updateNodes() {

        // Select an occupied node for transmission attempt
        Node node = sink;
        do {
            int randomIndex = random.nextInt(nodesWithPackets.size());
            node = nodesWithPackets.get(randomIndex);
        } while (node.equals(sink));

        ++numAttempts;

        Packet p = node.getPacket(0);

        Node prefNode = node;
        if (algo == 0)
            prefNode = algorithmA(node, p);
        else if (algo == 1)
            prefNode = algorithmB(node, p);

        if (!prefNode.equals(node)) {
            movePacket(node, prefNode, p);
        }
    }

    public List<Node> getNodes() {
        return nodes;
    }

    public List<Node> getNodesWithPackets() {
        return nodesWithPackets;
    }

    public Node getSink() {
        return sink;
    }

    public int getMaxQueue() {
        return maxQueue;
    }

    public int numMoves() {
        return numMoves;
    }

    public int numAttempts() {
        return numAttempts;
    }

    public int numPackets() {
        return numPackets;
    }

    public int numPacketsDelivered() {
        return sink.numPackets();
    }

    public static void exit(String s) {

        println(s);
        System.exit(0);
    }

    public static void println(String s) {
        System.out.println(s);
    }

    public String report() {

        String s = sink.numPackets() + " / " + numPackets
                + " packets delivered in " + numMoves + " moves / "
                + numAttempts + " attempts";
        println(s);
        return s;
    }

    public String summary() {

        String s = "\nSummary\n=======\n";
        s += report();
        println(s);
        return (s);
    }

    public boolean iterate() {

        for (int i = 0; i < 100; ++i) {

            updateNodes();
            report();
            if (sink.numPackets() == numPackets) {
                return true;
            }
        }
        return false;
    }

    public void go() {

        while (!iterate())
            ;

        summary();
    }

    public static void main(String[] args) {

        int width = 720;
        int height = 720;
        long randomSeed = 21122010;
        long configSeed = 32498572;
        int algo = 1;
        new SwarmNet(width, height, configSeed, randomSeed, algo).go();
    }
}
