package swarmnet;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class SwarmNet2 {

    private Network network;
    private List<Node> nodes, sourceNodes;
    private Random random;
    private Node sink;

    private int maxPackets, numPackets, maxQueue, numNodes, numMoves;
    private int numAttempts;

    private double neighbourDistance;

    private int width, height;

    public SwarmNet2(int w, int h, long configSeed, long randomSeed) {

        width = w;
        height = h;

        Random configRandom = new Random(configSeed);
        random = new Random(randomSeed);
        initNetwork(configRandom);

        makeSources();
        maxPackets = sourceNodes.size() * maxQueue * 10;
        numPackets = numMoves = numAttempts = 0;
    }

    private void initNetwork(Random configRandom) {

        maxQueue = 10;
        numNodes = 500;
        neighbourDistance = 50;
        
        network = new Network(width, height, numNodes, neighbourDistance,
                configRandom);
        
        nodes = network.getNodes();
        sink = network.getSink();

        sourceNodes = new ArrayList<Node>();
    }

    private void makeSources() {

        for (int i = 0; i < numNodes; ++i) {

            Node node = nodes.get(i);
            if ((node.getX() < width / 4 && node.getY() < width / 4)) {

                sourceNodes.add(node);
            }
        }

        if (sourceNodes.size() == 0)
            exit("Could not make any sources...exiting...");
    }

    private void generatePackets() {

        for (Node node : sourceNodes) {

            while (node.numPackets() < maxQueue) {

                Packet p = new Packet(node);
                node.addPacket(p);
                ++numPackets;
            }
        }
    }

    private void movePacket(Node fromNode, Node toNode, Packet p) {

        ++numMoves;

        p.setPreviousNode(fromNode);
        fromNode.removePacket(p);

        p.setNode(toNode);
        toNode.addPacket(p);
    }

    private Node randomNodeInList(List<Node> l) {
        return l.get(random.nextInt(l.size()));
    }

    // 1. If neighbour is sink, go there.
    // 2. Find most attractive available neighbour according to 
    // attraction = hopWeight * hops + mostAttractiveQueueLength - |deltaPackets|
    // where deltaPackets is difference between packets on neighbour and most attractive queue length
    // and hops is hop difference between this node and its neighbour i.e. hops = 1, 0 or -1.
    // If there is a tie, then the most attractive is chosen at random from the winning neighbours.
    // 3. If no neighbour is attractive, return to previous node, if this is available.
    private Node algorithm(Node node, Packet packet) {

        double hopWeight = 2.0;
        double mostAttrQ = maxQueue / 4.0;
        List<Node> attrNodes = new ArrayList<Node>();

        int currAttr = Integer.MIN_VALUE;
        List<Node> neighbours = node.getNeighbours();
        for (Node n : neighbours) {

            if (packet.getPreviousNode().equals(n))
                continue;

            if (n.equals(sink))
                return n;

            if (n.numPackets() == maxQueue)
                continue;

            int hops = node.getHopsToSink() - n.getHopsToSink();
            int attr = (int) Math.round(hopWeight * hops 
                    - Math.abs(n.numPackets() - mostAttrQ));

            if (attr == currAttr)
                attrNodes.add(n);

            else if (attr > currAttr) {
                attrNodes.clear();
                attrNodes.add(n);
                currAttr = attr;
            }
        }

        if (!attrNodes.isEmpty())
            return randomNodeInList(attrNodes);
        else if (packet.getPreviousNode().numPackets() < maxQueue)
            return packet.getPreviousNode();

        return node;
    }

    // From empirical tests using the ns-2 network simulator:
    // max Q length = 50
    // pks generated at 25, 35 or 45 pks/sec from a few active nodes in close proximity.
    // Measure packet delivery rate (PDR) and end to end delay (EED).
    private void updateNodes() {

        // Select an occupied node for transmission attempt
        Node node = sink;
        do {
            int randomIndex = random.nextInt(nodes.size());
            node = nodes.get(randomIndex);
        } while (node.equals(sink) || node.isEmpty());

        ++numAttempts;

        Packet p = node.getPacket(0);

        Node prefNode = node;

        prefNode = algorithm(node, p);

        if (!prefNode.equals(node)) {
            movePacket(node, prefNode, p);
        }
    }

    public List<Node> getNodes() {
        return nodes;
    }

    public Node getSink() {
        return sink;
    }

    public int getMaxQueue() {
        return maxQueue;
    }

    public int numMoves() {
        return numMoves;
    }

    public int numAttempts() {
        return numAttempts;
    }

    public int numPackets() {
        return maxPackets;
    }

    public int numPacketsDelivered() {
        return sink.numPackets();
    }

    public static void exit(String s) {

        println(s);
        System.exit(0);
    }

    public static void println(String s) {
        System.out.println(s);
    }

    public String report() {

        String s = sink.numPackets() + " / " + maxPackets
                + " packets delivered in " + numMoves + " moves / "
                + numAttempts + " attempts";
        println(s);
        return s;
    }

    public String summary() {

        String s = "\nSummary\n=======\n";
        s += report();
        println(s);
        return (s);
    }

    public boolean iterate() {

        generatePackets();

        for (int i = 0; i < 100; ++i) {

            updateNodes();
            report();
            if (sink.numPackets() >= maxPackets) {
                return true;
            }
        }
        return false;
    }

    public void go() {

        while (!iterate())
            ;

        summary();
    }

    public static void main(String[] args) {

        int width = 720;
        int height = 720;
        long randomSeed = 21122010;
        long configSeed = 32498572;
        new SwarmNet2(width, height, configSeed, randomSeed).go();
    }
}
